
#******************************************************************************
#                           tplsh.tcl   (TP-Link SmartHome)
#******************************************************************************
#.... Author              : Marquard, Thomas - alias: HM42
#.... Version             : see: scriptVersInfo
#.... Comment             : TCL script for data exchange between Homematic CCUs
#                           and TP-Link Wi-Fi Smart Home devices,
#                           e.g.: HS100 or HS110
#
#.... History             : 2018-11-25, Vers. 1.0, basic version
#
#                           2018-12-03, Vers. 2.0, change from button to 
#                             switch mode; reading ("readdata") of current
#                             voltage, power, energy values from 
#                             HS110, HW version 1
#
#                           2019-01-01, Vers. 3.0, extension for
#                             HS110, HW version 2
#
#                           Based on Python script from Lubomir Stroetmann
#                           Copyright 2016 softScheck GmbH
#                           see: https://www.softscheck.com/en/
#                                  reverse-engineering-tp-link-hs110/
#
#                           Licensed under the Apache License, 
#                           Version 2.0 (the "License");
#                           you may not use this file except in compliance 
#                           with the License. 
#
#                           You may obtain a copy of the License at
#
#                           http://www.apache.org/licenses/LICENSE-2.0
#
#                           Unless required by applicable law or agreed to 
#                           in writing, software distributed under the License
#                           is distributed on an "AS IS" BASIS,
#                           WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
#                           either express or implied.
#                           See the License for the specific language governing 
#                           permissions and limitations under the License.
#
#******************************************************************************

load tclrega.so

global scriptVersInfo
global copyRight
global env

set versInfo  {Script:    'tplsh.tcl', Version: 3.0, 2019-01-01}
set copyright {Copyright: Thomas Marquard / Lubomir Stroetmann, softScheck GmbH (...2019)}


exec clear >@ stdout 
puts ""
puts $versInfo
puts $copyright
puts ""


# Predefined Smart Plug Commands

set commandArray {
  { info      \{"system":{"get_sysinfo":{}}\} }
  { on        \{"system":{"set_relay_state":{"state":1}}\} }
  { off       \{"system":{"set_relay_state":{"state":0}}\} }
  { toggle    \{"system":{"get_sysinfo":{}}\} }
  { readdata  \{"emeter":{"get_realtime":{}}\} }
}



proc usagePuts {} {
 
  puts ""
  puts "USAGE: tclsh /usr/local/etc/config/addons/tplsh/tplsh.tcl <hostname or ip-address> <command>"
  puts ""
  puts "COMMAND: \[info|on|off|toggle|readdata\]"
  puts ""

}   ;# usagePuts



proc handleErrorMessage {message} {

  set regaCmd "
		object objVar = dom.GetObject(ID_SYSTEM_VARIABLES).Get(\"FehlerMeldung\");
		if (!objVar) \{
			object objSysVarIDs = dom.GetObject(ID_SYSTEM_VARIABLES);
			objVar = dom.CreateObject(OT_VARDP);
			objSysVarIDs.Add(objVar.ID());
			objVar.Name(\"FehlerMeldung\");
			objVar.ValueUnit(\"\");
			objVar.DPInfo(\"\");
			objVar.ValueType(ivtString);
			objVar.ValueSubType(istChar8859);
			dom.RTUpdate(false);
		\}
		if (dom.GetObject(\"FehlerMeldung\").State() <> \"tplsh.tcl, $message\") 
		\{
			dom.GetObject(\"FehlerMeldung\").State(\"tplsh.tcl, $message\");
		\}      
  "

  rega_script $regaCmd

}   ;# handleErrorMessage



# Encryption and Decryption of TP-Link Smart Home Protocol
# XOR Autokey Cipher with starting key = 171

proc encrypt {message} {

  set key 171
  set result [binary format I [string length $message]]
  
  for {set charIdx 0} {$charIdx < [string length $message]} {incr charIdx} {
    set char [string index $message $charIdx]
    scan $char %c ascii
    set a [expr $key ^ $ascii] 
    set key $a
    append result [binary format c $a] 
  } 

  return $result
  
}   ;# encrypt  



proc decrypt {message} {

  set key 171
  set result ""

  for {set charIdx 4} {$charIdx < [string length $message]} {incr charIdx} {
    set char [string index $message $charIdx]
    scan $char %c ascii
    set a [expr $key ^ $ascii] 
    set key $ascii
    append result [binary format c $a] 
  } 

  return $result
  
}   ;# decrypt  



########## MAIN SCRIPT ######################################################################


if {$argc >= 2} {

  set hostnameOrIpAddress [lindex $argv 0]
  set command [lindex $argv 1]
  
} else {

  set msg "ERROR: One or two program parameters missing !"
  handleErrorMessage $msg
  puts ""
  puts $msg
  puts ""
  usagePuts
  exit
}

foreach {one} $commandArray {
  lappend commands [lindex $one 0]
}
set commandIdx [lsearch $commands $command] 

if {$commandIdx < 0} {

  set msg "ERROR: Unknown command `$command` !"
  handleErrorMessage $msg
  puts ""
  puts $msg
  puts ""
  usagePuts
  exit
}

set port 9999

if {$command == "toggle" || $command == "readdata"} {

	if { [catch {set sock [socket $hostnameOrIpAddress $port]}] } {
		set msg "ERROR: Cannot open port `$port` at `$hostnameOrIpAddress` !"
		handleErrorMessage $msg
		puts ""
		puts $msg
    puts ""
		usagePuts
		exit
	}

	fconfigure $sock -encoding binary -blocking 0

  set commandIdx [lsearch $commands info] 
  set commandString [lindex [lindex $commandArray $commandIdx] 1]
  set encryptedCommandString [encrypt $commandString]
  puts -nonewline $sock $encryptedCommandString 
  flush $sock

  set answer ""
  for {set waitIdx 0} {$waitIdx < 20} {incr waitIdx} {
    append answer [read $sock]
    if { [string length $answer] > 4 } {
      break
    } 
    set vwaitEnd 0
    after 50 set vwaitEnd 1
    vwait vwaitEnd
  }

  close $sock

  set answer [decrypt $answer]
  puts "Info message received: $answer"

  set startIdx [expr [string first \"relay_state\": $answer] + 1]
  set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
  set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
  set relayState [string range $answer $startValueIdx $endValueIdx]


  if {$command == "toggle"} {	
		switch $relayState {

			"0" {
				set command on
			} 

			"1" {
				set command off
			} 

			default {
				set msg "ERROR: Unknown actual relay state read by toggle command: `$relayState` !"
				handleErrorMessage $msg
				puts ""
				puts $msg
				puts ""
				usagePuts
				catch {close $sock}
				exit
			}
		}
	}
}


if { [catch {set sock [socket $hostnameOrIpAddress $port]}] } {
  set msg "ERROR: Cannot open port `$port` at `$hostnameOrIpAddress` !"
  handleErrorMessage $msg
  puts ""
  puts $msg
  puts ""
  usagePuts
  exit
}

fconfigure $sock -encoding binary -blocking 0

set commandIdx [lsearch $commands $command] 
set commandString [lindex [lindex $commandArray $commandIdx] 1]
set encryptedCommandString [encrypt $commandString]
  
puts -nonewline $sock $encryptedCommandString 
flush $sock

puts ""
puts "Message sent: $commandString"
puts ""

set answer ""
for {set waitIdx 0} {$waitIdx < 20} {incr waitIdx} {
  append answer [read $sock]
  if { [string length $answer] > 4 } {
    break
  } 
  set vwaitEnd 0
  after 50 set vwaitEnd 1
  vwait vwaitEnd
}

set answer [decrypt $answer]
puts "Message received: $answer"
puts ""

close $sock


if {$command == "readdata"} {
	if {[catch {set sourceChnName $::env(CUXD_CHANNEL)}]} { 
		set msg "ERROR: No channel specified in environment variable `CUXD_CHANNEL` (script not called from CUxD) !"
		handleErrorMessage $msg
		puts ""
		puts $msg
		puts ""
		exit
	}

	set regaCmd "
		dom.GetObject(\"CUxD.${sourceChnName}.SET_STATE\").State($relayState);
  "

	rega_script $regaCmd


	set startIdx [expr [string first \"current\": $answer] + 1]
	
	if {$startIdx > 0} {
		# readdata for HS110, HW version 1 ...
		set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
		set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
		set current [format "%.2f" [string range $answer $startValueIdx $endValueIdx]]

		set startIdx [expr [string first \"voltage\": $answer] + 1]
		set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
		set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
		set voltage [format "%.2f" [string range $answer $startValueIdx $endValueIdx]]

		set startIdx [expr [string first \"power\": $answer] + 1]
		set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
		set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
		set power [format "%.2f" [string range $answer $startValueIdx $endValueIdx]]

		set startIdx [expr [string first \"total\": $answer] + 1]
		set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
		set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
		set energy [format "%.3f" [string range $answer $startValueIdx $endValueIdx]]
	} else {
		set startIdx [expr [string first \"current_ma\": $answer] + 1]
		if {$startIdx > 0} {
      # readdata for HS110, HW version 2 ...
			set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
			set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
			set current [format "%.2f" [expr [string range $answer $startValueIdx $endValueIdx] / 1000.]]

			set startIdx [expr [string first \"voltage_mv\": $answer] + 1]
			set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
			set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
			set voltage [format "%.2f" [expr [string range $answer $startValueIdx $endValueIdx] / 1000.]]

			set startIdx [expr [string first \"power_mw\": $answer] + 1]
			set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
			set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
			set power [format "%.2f" [expr [string range $answer $startValueIdx $endValueIdx] / 1000.]]

			set startIdx [expr [string first \"total_wh\": $answer] + 1]
			set startValueIdx [expr [string first ":" $answer $startIdx] + 1]
			set endValueIdx [expr [string first "," $answer $startValueIdx] - 1]
			set energy [format "%.3f" [expr [string range $answer $startValueIdx $endValueIdx] / 1000.]]
		} else {
      exit;
 		}		
	}	
		
	puts "Voltage: $voltage V / Current: $current A / Power: $power W / Energy: $energy Wh"
	puts ""
	
	set regaCmd "
		string strQuellKanalBez;
		object objDP = dom.GetObject(\"CUxD.${sourceChnName}.CMD_RETS\");
		strQuellKanalBez = (dom.GetObject(objDP.Channel())).Name();
		
		object objVar = dom.GetObject(ID_SYSTEM_VARIABLES).Get(\"Strom_\" # strQuellKanalBez);
		if (!objVar) \{
			object objSysVarIDs = dom.GetObject(ID_SYSTEM_VARIABLES);
			objVar = dom.CreateObject(OT_VARDP);
			objSysVarIDs.Add(objVar.ID());
			objVar.Name(\"Strom_\" # strQuellKanalBez);
			objVar.ValueUnit(\"A\");
			objVar.ValueType(ivtFloat);
			objVar.ValueSubType(istGeneric);
			objVar.DPInfo(\"tp-Link Steckdose\");
			objVar.ValueMin(0);
			objVar.ValueMax(20);
			objVar.State(0);
			dom.RTUpdate(false);
		\}
		objVar.State($current);

		object objVar = dom.GetObject(ID_SYSTEM_VARIABLES).Get(\"Spannung_\" # strQuellKanalBez);
		if (!objVar) \{
			object objSysVarIDs = dom.GetObject(ID_SYSTEM_VARIABLES);
			objVar = dom.CreateObject(OT_VARDP);
			objSysVarIDs.Add(objVar.ID());
			objVar.Name(\"Spannung_\" # strQuellKanalBez);
			objVar.ValueUnit(\"V\");
			objVar.ValueType(ivtFloat);
			objVar.ValueSubType(istGeneric);
			objVar.DPInfo(\"tp-Link Steckdose\");
			objVar.ValueMin(0);
			objVar.ValueMax(250);
			objVar.State(0);
			dom.RTUpdate(false);
		\}
		objVar.State($voltage);
		
		object objVar = dom.GetObject(ID_SYSTEM_VARIABLES).Get(\"Leistung_\" # strQuellKanalBez);
		if (!objVar) \{
			object objSysVarIDs = dom.GetObject(ID_SYSTEM_VARIABLES);
			objVar = dom.CreateObject(OT_VARDP);
			objSysVarIDs.Add(objVar.ID());
			objVar.Name(\"Leistung_\" # strQuellKanalBez);
			objVar.ValueUnit(\"W\");
			objVar.ValueType(ivtFloat);
			objVar.ValueSubType(istGeneric);
			objVar.DPInfo(\"tp-Link Steckdose\");
			objVar.ValueMin(0);
			objVar.ValueMax(100000);
			objVar.State(0);
			dom.RTUpdate(false);
		\}
		objVar.State($power);
		
		object objVar = dom.GetObject(ID_SYSTEM_VARIABLES).Get(\"Energie_\" # strQuellKanalBez);
		if (!objVar) \{
			object objSysVarIDs = dom.GetObject(ID_SYSTEM_VARIABLES);
			objVar = dom.CreateObject(OT_VARDP);
			objSysVarIDs.Add(objVar.ID());
			objVar.Name(\"Energie_\" # strQuellKanalBez);
			objVar.ValueUnit(\"kWh\");
			objVar.ValueType(ivtFloat);
			objVar.ValueSubType(istGeneric);
			objVar.DPInfo(\"tp-Link Steckdose\");
			objVar.ValueMin(0);
			objVar.ValueMax(100000);
			objVar.State(0);
			dom.RTUpdate(false);
		\}
		objVar.State($energy);
	
	"
	
	rega_script $regaCmd
  
}
	
# end